import os
import subprocess
import sys
import tempfile
import tomllib
import unittest
from io import StringIO
from unittest.mock import Mock, patch

from .config import DotDict, get_latest_semantic_version_tag, parse_semantic_version


class TestEnhancedDotDict(unittest.TestCase):
    """Test cases for the enhanced DotDict class with TOML support."""

    def setUp(self):
        """Initialize test data and capture stdout for warning messages."""
        self.sample_dict = {
            "server": {"host": "localhost", "port": 8080},
            "database": {
                "credentials": {"username": "admin", "password": "secret123"},
                "settings": {"timeout": 30},
            },
            "users": [{"name": "Alice", "role": "admin"}, {"name": "Bob", "role": "user"}],
            "valid_key": "value",
            "invalid-key": "problematic",
            "123numeric": "starts with number",
            "get": "overrides method",
        }

        # Valid TOML content
        self.valid_toml = """
[server]
host = "localhost"
port = 8080

[database.credentials]
username = "admin"
password = "secret123"

[database.settings]
timeout = 30

[[users]]
name = "Alice"
role = "admin"

[[users]]
name = "Bob"
role = "user"

valid_key = "value"
invalid-key = "problematic"
123numeric = "starts with number"
get = "overrides method"
"""

        # Invalid TOML content (syntax error)
        self.invalid_toml = """
[server]
host = "localhost
port = 8080
"""

        # Create temporary files for TOML testing
        self.valid_toml_file = tempfile.NamedTemporaryFile(  # noqa: SIM115
            delete=False, mode="w", suffix=".toml"
        )
        self.valid_toml_file.write(self.valid_toml)
        self.valid_toml_file.close()

        self.invalid_toml_file = tempfile.NamedTemporaryFile(  # noqa: SIM115
            delete=False, mode="w", suffix=".toml"
        )
        self.invalid_toml_file.write(self.invalid_toml)
        self.invalid_toml_file.close()

        # Setup stdout capture
        self.captured_output = StringIO()
        self.original_stdout = sys.stdout
        sys.stdout = self.captured_output

    def tearDown(self):
        """Clean up temporary files and restore stdout."""
        os.unlink(self.valid_toml_file.name)
        os.unlink(self.invalid_toml_file.name)
        sys.stdout = self.original_stdout

    def test_init_with_dict(self):
        """Test initialization with a dictionary."""
        dot_dict = DotDict(self.sample_dict)

        # Test basic access
        self.assertEqual(dot_dict.server.host, "localhost")
        self.assertEqual(dot_dict.server.port, 8080)
        self.assertEqual(dot_dict.database.credentials.username, "admin")

        # Test list of dicts
        self.assertEqual(dot_dict.users[0].name, "Alice")
        self.assertEqual(dot_dict.users[1].role, "user")

        # Verify warnings were printed
        output = self.captured_output.getvalue()
        self.assertIn("Warning: 'invalid-key'", output)
        self.assertIn("Warning: '123numeric'", output)
        self.assertIn("Warning: 'get'", output)

    def test_load_valid_toml_file(self):
        """Test loading a valid TOML file."""
        # Temporarily reset stdout to avoid capturing import warnings
        sys.stdout = self.original_stdout

        # Restore stdout capture
        sys.stdout = self.captured_output

        # Load TOML file
        dot_dict = DotDict.from_toml_file(self.valid_toml_file.name)

        # Test basic access
        self.assertEqual(dot_dict.server.host, "localhost")
        self.assertEqual(dot_dict.server.port, 8080)
        self.assertEqual(dot_dict.database.credentials.username, "admin")

        # Test list of dicts
        self.assertEqual(dot_dict.users[0].name, "Alice")
        self.assertEqual(dot_dict.users[1].role, "user")

        # Verify warnings were printed
        output = self.captured_output.getvalue()
        self.assertIn("Warning: 'invalid-key'", output)
        self.assertIn("Warning: '123numeric'", output)
        self.assertIn("Warning: 'get'", output)

    def test_load_invalid_toml_file(self):
        """Test loading an invalid TOML file raises appropriate exception."""
        # Reset and restore stdout to avoid capturing import warnings
        sys.stdout = self.original_stdout
        sys.stdout = self.captured_output

        # Try to load invalid TOML file
        with self.assertRaises(tomllib.TOMLDecodeError):
            DotDict.from_toml_file(self.invalid_toml_file.name)

    def test_access_invalid_keys(self):
        """Test accessing keys that are not valid Python identifiers."""
        dot_dict = DotDict(self.sample_dict)

        # These should work with dictionary access
        self.assertEqual(dot_dict["invalid-key"], "problematic")
        self.assertEqual(dot_dict["123numeric"], "starts with number")
        self.assertEqual(dot_dict["get"], "overrides method")

        # But fail with attribute access
        with self.assertRaises(AttributeError):
            _ = dot_dict.invalid - key  # noqa

        # This might work in Python, but it's a subtraction operation
        # dot_dict.invalid-key is interpreted as (dot_dict.invalid) - key
        with self.assertRaises(AttributeError):
            _ = dot_dict.invalid

        # with self.assertRaises(NameError):
        #     _ = dot_dict.123numeric  # SyntaxError at compile time

    def test_deep_nesting(self):
        """Test deeply nested structures."""
        deep_dict = {"level1": {"level2": {"level3": {"level4": {"value": "deep"}}}}}

        dot_dict = DotDict(deep_dict)
        self.assertEqual(dot_dict.level1.level2.level3.level4.value, "deep")

    def test_update_nested_attribute(self):
        """Test updating a nested attribute."""
        dot_dict = DotDict(self.sample_dict)

        # Update nested attribute
        dot_dict.server.host = "new-host"
        self.assertEqual(dot_dict.server.host, "new-host")
        self.assertEqual(dot_dict["server"]["host"], "new-host")

        # Update with a new dictionary should convert to DotDict
        dot_dict.server = {"host": "another-host", "new_key": "value"}
        self.assertEqual(dot_dict.server.host, "another-host")
        self.assertEqual(dot_dict.server.new_key, "value")

        # Verify it's a DotDict
        self.assertIsInstance(dot_dict.server, DotDict)

    def test_from_toml(self):
        """Test creating DotDict from already parsed TOML data."""
        try:
            # Reset and restore stdout to avoid capturing import warnings
            sys.stdout = self.original_stdout
            with open(self.valid_toml_file.name, "rb") as f:
                toml_data = tomllib.load(f)

            sys.stdout = self.captured_output

            # Create DotDict from parsed TOML
            dot_dict = DotDict.from_toml_data(toml_data)

            # Test access
            self.assertEqual(dot_dict.server.host, "localhost")
            self.assertEqual(dot_dict.users[0].name, "Alice")

        except ImportError:
            self.skipTest("TOML library not available")

    def test_nested_list_handling(self):
        """Test handling of nested lists with dictionaries."""
        complex_dict = {
            "nested_lists": [
                [{"name": "item1"}, {"name": "item2"}],
                [{"name": "item3"}, {"name": "item4"}],
            ]
        }

        dot_dict = DotDict(complex_dict)

        # Test that dictionaries in nested lists are converted
        self.assertEqual(dot_dict.nested_lists[0][0].name, "item1")
        self.assertEqual(dot_dict.nested_lists[1][1].name, "item4")

        # Verify outer list items are DotDict instances
        self.assertIsInstance(dot_dict.nested_lists[0][0], DotDict)


class TestSemanticVersionTag(unittest.TestCase):

    @patch("subprocess.run")
    def test_get_latest_semantic_version_tag_valid(self, mock_run):
        mock_process = Mock()
        mock_process.stdout = "v1.0.0\nv0.9.0\nv1.0.1\nother_tag\n"
        mock_run.return_value = mock_process

        latest_tag = get_latest_semantic_version_tag()
        self.assertEqual(latest_tag, "v1.0.1")

    @patch("subprocess.run")
    def test_get_latest_semantic_version_tag_no_semantic_tags(self, mock_run):
        mock_process = Mock()
        mock_process.stdout = "tag1\ntag2\ntag3\n"
        mock_run.return_value = mock_process

        latest_tag = get_latest_semantic_version_tag()
        self.assertIsNone(latest_tag)

    @patch("subprocess.run")
    def test_get_latest_semantic_version_tag_empty_tags(self, mock_run):
        mock_process = Mock()
        mock_process.stdout = ""
        mock_run.return_value = mock_process

        latest_tag = get_latest_semantic_version_tag()
        self.assertIsNone(latest_tag)

    @patch("subprocess.run")
    def test_get_latest_semantic_version_tag_git_error(self, mock_run):
        mock_run.side_effect = subprocess.CalledProcessError(1, "git tag")
        latest_tag = get_latest_semantic_version_tag()
        self.assertIsNone(latest_tag)

    @patch("subprocess.run")
    def test_get_latest_semantic_version_tag_file_not_found(self, mock_run):
        mock_run.side_effect = FileNotFoundError()
        latest_tag = get_latest_semantic_version_tag()
        self.assertIsNone(latest_tag)

    def test_parse_semantic_version_valid(self):
        self.assertEqual(parse_semantic_version("v1.2.3"), (1, 2, 3))
        self.assertEqual(parse_semantic_version("v0.0.1"), (0, 0, 1))

    def test_parse_semantic_version_invalid(self):
        self.assertIsNone(parse_semantic_version("1.2.3"))
        self.assertIsNone(parse_semantic_version("v1.2"))
        self.assertIsNone(parse_semantic_version("v1.2.3-alpha"))
        self.assertIsNone(parse_semantic_version("abc"))


if __name__ == "__main__":
    unittest.main()
